import { DependencyGraph } from "../dependency-graph/dependency-graph";
import { Builder } from "./builder/builder";
import { ValidationTypes } from "../validators/validation-type";
import { RuleSet } from "./rule-set/rule-set";
export class RuleEngine {
    constructor(options) {
        this.propertyMap = {};
        this.dependencyGraph = new DependencyGraph();
        this.validations = new WeakMap();
        this.dataLinks = new Map();
        this.snapshots = new Map();
        this.propertiesThatRequireAnEagerUpdate = [];
        this.builder = new Builder(options, this, this.dependencyGraph, this.propertyMap);
    }
    get properties() {
        return Object.values(this.propertyMap);
    }
    defineRuleSet(initFcn) {
        return new RuleSet(initFcn, this.builder);
    }
    getBuilder() {
        return this.builder;
    }
    getPropertyById(id) {
        return this.propertyMap[id];
    }
    takeSnapShot(key = 'default') {
        const snap = {
            key,
            data: this.properties.map(p => p.exportData())
        };
        this.snapshots.set(key, snap);
        return snap;
    }
    restoreSnapShot(key = 'default') {
        const snap = this.snapshots.get(key);
        if (snap) {
            snap.data.forEach((d, i) => this.properties[i].importData(d));
        }
    }
    linkPropertyData(propertyA, propertyB) {
        const dataLinkKey = this.createDataLinkKey(propertyA, propertyB);
        if (this.dataLinks.has(dataLinkKey)) {
            return;
        }
        const listenerA = {
            updated: () => propertyA.transferData(propertyB)
        };
        const listenerB = {
            updated: () => propertyB.transferData(propertyA)
        };
        propertyA.transferData(propertyB);
        propertyA.registerValueChangedListener(listenerA);
        propertyB.registerValueChangedListener(listenerB);
        this.dataLinks.set(dataLinkKey, [listenerA, listenerB]);
    }
    unlinkPropertyData(propertyA, propertyB) {
        const dataLinkKey = this.createDataLinkKey(propertyA, propertyB);
        const listeners = this.dataLinks.get(dataLinkKey);
        if (listeners) {
            const [listenerA, listenerB] = listeners;
            propertyA.deregisterValueChangedListener(listenerA);
            propertyB.deregisterValueChangedListener(listenerB);
            this.dataLinks.delete(dataLinkKey);
        }
    }
    createDataLinkKey(propertyA, propertyB) {
        return propertyA.id < propertyB.id
            ? `${propertyA.id}<->${propertyB.id}`
            : `${propertyB.id}<->${propertyA.id}`;
    }
    needsAnUpdate(mightHaveChanged) {
        this.dependencyGraph.traverseDepthFirst(mightHaveChanged.id, property => {
            property.needsAnUpdate(false);
            if (property.internallyRequiresEagerUpdate()
                && this.propertiesThatRequireAnEagerUpdate.every(p => p.id !== property.id)) {
                this.propertiesThatRequireAnEagerUpdate.push(property);
            }
        }, dependency => !!dependency.options.value);
        setTimeout(() => {
            const eagerProps = this.propertiesThatRequireAnEagerUpdate;
            this.propertiesThatRequireAnEagerUpdate = [];
            eagerProps.forEach((prop) => void this.updateValue(prop));
        }, 0);
    }
    updateValue(property) {
        var _a;
        const asyncUpdates = (_a = this.dependencyGraph.getAsyncDependencies(property.id)) === null || _a === void 0 ? void 0 : _a.map(asyncDep => this.updateValue(asyncDep));
        if ((asyncUpdates === null || asyncUpdates === void 0 ? void 0 : asyncUpdates.length) && asyncUpdates.some(update => !!update)) {
            return Promise.all(asyncUpdates).then(() => this.updateValueInternal(property)).catch((err) => {
                property.errorWhileUpdating(err);
            });
        }
        else {
            return this.updateValueInternal(property);
        }
    }
    updateValueInternal(property) {
        try {
            const updatePromise = property.internallyUpdate();
            if (updatePromise) {
                return updatePromise.then(() => {
                    this.hasBeenUpdated(property);
                }).catch(err => property.errorWhileUpdating(err));
            }
            else {
                this.hasBeenUpdated(property);
            }
        }
        catch (err) {
            property.errorWhileUpdating(err);
        }
    }
    hasBeenUpdated(property) {
        property.hasBeenUpdated();
        this.dependencyGraph.traverseDepthFirst(property.id, (prop, dep) => prop.dependencyHasBeenUpdated(dep), dependency => !dependency.options.value, true);
    }
    invalidateValidationResults(validators) {
        validators.forEach(v => {
            const vprocess = this.getValidationProcess(v);
            vprocess.isLastResultUpToDate = false;
        });
    }
    validate(validators) {
        return validators.map(validator => {
            const vprocess = this.getValidationProcess(validator);
            if (vprocess.isLastResultUpToDate) {
                return Promise.resolve(vprocess.lastValidationResult);
            }
            else if (vprocess.currentValidation) {
                return vprocess.currentValidation;
            }
            const validation = validator.validate(...validator.getValidatedProperties());
            if (validation instanceof Promise) {
                vprocess.currentValidation = validation.then(res => {
                    vprocess.currentValidation = undefined;
                    vprocess.lastValidationResult = res;
                    vprocess.isLastResultUpToDate = true;
                    return res;
                }, err => {
                    vprocess.currentValidation = undefined;
                    return [{
                            text: JSON.stringify(err),
                            type: ValidationTypes.ErrorThrownDuringValidation
                        }];
                });
                return vprocess.currentValidation;
            }
            else {
                vprocess.lastValidationResult = validation;
                vprocess.isLastResultUpToDate = true;
                return Promise.resolve(validation);
            }
        });
    }
    getValidationProcess(validator) {
        const vprocess = this.validations.get(validator);
        if (vprocess) {
            return vprocess;
        }
        else {
            const newProcess = {
                isLastResultUpToDate: false
            };
            this.validations.set(validator, newProcess);
            return newProcess;
        }
    }
}
